from pathlib import Path
from radiens.grpc import (allegoserver_pb2, allegoserver_pb2_grpc,
                          common_pb2, spikesorter_pb2)

DEFAULT_IP = 'localhost'
DEFAULT_HUB = 'default'


# ports
ALLEGO_CORE_PORT = 50051
PCACHE_PORT = 50052
KPI_PORT = 50053
NEURONS1_PORT = 50054

# default server addresses
CORE_ADDR = '{}:{}'.format(DEFAULT_IP, ALLEGO_CORE_PORT)
PCACHE_ADDR = '{}:{}'.format(DEFAULT_IP, PCACHE_PORT)
KPI_ADDR = '{}:{}'.format(DEFAULT_IP, KPI_PORT)
NEURONS1_ADDR = '{}:{}'.format(DEFAULT_IP, NEURONS1_PORT)


# services
CORE_SERVICE = 'core'
PCACHE_SERVICE = 'pcache'

# only allego stream group
PRIMARY_CACHE_STREAM_GROUP_ID = 'Live Signals'

# default datasource paths
DATASOURCE_PATHS = {
    'dsrc_path': Path(Path.home(), 'radix', 'data'),
    'dsrc_type': 'xdat',
    'dsrc_name': 'sbpro'
}

# ports/signals
PORT_ENUM = {0: 'A', 1: 'B', 2: 'C', 3: 'D', 4: 'E', 5: 'F', 6: 'G', 7: 'H'}
PORT_ = {'A': common_pb2.A,  'B': common_pb2.B, 'C': common_pb2.C,  'D': common_pb2.D,
         'E': common_pb2.E,  'F': common_pb2.F, 'G': common_pb2.G,  'H': common_pb2.H}
SIGNAL_TYPE_ENUM = {0: 'amp', 1: 'adc', 2: 'din', 3: 'dout'}
SIGNAL_TYPES = ['amp', 'adc', 'din', 'dout']

# system
SYSTEM_MODE = {
    'sbpro': allegoserver_pb2.SMARTBOX_PRO,
    'sbpro-sinaps-256': allegoserver_pb2.SMARTBOX_PRO_SINAPS_256,
    'sbpro-sinaps-1024': allegoserver_pb2.SMARTBOX_PRO_SINAPS_1024,
    'sb-classic': allegoserver_pb2.SMARTBOX_CLASSIC,
    'sim-sine': allegoserver_pb2.SMARTBOX_SIM_GEN_SINE,
    'sim-sine-mapped': allegoserver_pb2.SMARTBOX_SIM_GEN_SINE_MAPPED,
    'sim-sine-high-freq': allegoserver_pb2.SMARTBOX_SIM_GEN_SINE_HIGH_FREQ,
    'sim-sine-multi-band': allegoserver_pb2.SMARTBOX_SIM_GEN_SINE_MULTI_BAND,
    'sim-spikes': allegoserver_pb2.SMARTBOX_SIM_GEN_SPIKES,
    'open-ephys_usb2': allegoserver_pb2.OPEN_EPHYS_USB2,
    'open-ephys_usb3': allegoserver_pb2.OPEN_EPHYS_USB3,
    'intan-usb2': allegoserver_pb2.INTAN_USB2,
    'intan-1024': allegoserver_pb2.INTAN_RECORDING_CONTROLLER_1024,
    'intan-512': allegoserver_pb2.INTAN_RECORDING_CONTROLLER_512,
    'xdaq-one-rec': allegoserver_pb2.XDAQ_ONE_REC,
    'xdaq-one-stim': allegoserver_pb2.XDAQ_ONE_STIM,
    'xdaq-core-rec': allegoserver_pb2.XDAQ_CORE_REC,
    'xdaq-core-stim': allegoserver_pb2.XDAQ_CORE_STIM,
}

SYSTEM_MODE_DECODE = {
    0: 'sbox-pro',
    1: 'sim-sine',
    2: 'sim-spikes',
    4: 'open-ephys-usb3',
    5: 'intan-1024',
    6: 'sb-classic',
    7: 'intan-512',
    8: 'open-ephys-usb2',
    9: 'intan-usb2',
    10: 'sim-sine-mapped',
    11: 'sim-sine-high-freq',
    12: 'sim-sine-multi-band',
    13: 'sbpro-sinaps-256',
    14: 'sbpro-sinaps-1024',
    15: 'xdaq-one-rec',
    16: 'xdaq-one-stim',
    17: 'xdaq-core-rec',
    18: 'xdaq-core-stim',
}

DOUT_MODE_OUT = {
    0: 'manual',
    1: 'events',
    2: 'gated'
}

APPS_ = {0: 'allego', 1: 'curate', 2: 'videre'}

# stream/record
STREAM_MODE = {'S_OFF': 0, 'S_ON': 1}
RECORD_MODE = {'R_OFF': 0, 'R_ON': 1}
STREAM_MODE_OUT = {0: 'S_OFF', 1: 'S_ON'}
RECORD_MODE_OUT = {0: 'R_OFF', 1: 'R_ON'}

# sensors
HEADSTAGE_ALIAS = {'smart-16': 'acute_smartlink_A16', 'smart-32': 'chronic_smartlink_CM32',
                   'pass-32': 'passthrough_32',  'pass-64': 'passthrough_64', 'chronic-64': 'chronic_smartlink_H64'}

PROBE_ALIAS = {'v1x16-16': 'v1x16_edge_10mm200_177', '4x8-32': 'a4x8_5mm100_200_177',
               '8x1tet-32': 'a8x1_tet_7mm_200_121',  'buz-32': 'buz32', 'poly3-32': 'a1x32_poly3_8mm50s_177',
               ' poly5-32': 'a1x32_poly5_6mm35s_100',  '8x8-64': 'a8x8_5mm200_200_413', 'poly3-64': 'v1x64_poly3_10mm25s_177',
               'buz-64': 'buz64'}

# spike sorter
SPK_SORTER_ = {'on': spikesorter_pb2.SORTER_CMD_ON, 'off': spikesorter_pb2.SORTER_CMD_OFF,
               'clear': spikesorter_pb2.SORTER_CMD_CLEAR_SORT, 'rebase': spikesorter_pb2.SORTER_CMD_REBASE,
               'init': spikesorter_pb2.SORTER_CMD_INIT, 'localize': spikesorter_pb2.SORTER_CMD_LOCALIZE,
               'sort': spikesorter_pb2.SORTER_CMD_SORT}

SPK_SORTER_STATE_DECODE = {0: 'on', 1: 'off', 2: 'not_configured'}



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: spikesorter.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import common_pb2 as common__pb2
from . import biointerface_pb2 as biointerface__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x11spikesorter.proto\x12\x06\x61llego\x1a\x0c\x63ommon.proto\x1a\x12\x62iointerface.proto\"\x84\x02\n\x18SpikeSorterLaunchRequest\x12\x11\n\tdsourceID\x18\x01 \x01(\t\x12&\n\x06source\x18\x02 \x01(\x0b\x32\x16.allego.FileDescriptor\x12\x1b\n\x13maxNumSiteNeighbors\x18\x04 \x01(\x05\x12\x18\n\x10neighborRadiusUm\x18\x05 \x01(\x01\x12/\n\x07variant\x18\x06 \x03(\x0b\x32\x1e.allego.SpikeSorterVariantSpec\x12%\n\x04mode\x18\x07 \x01(\x0e\x32\x17.allego.SpikeSorterMode\x12\x0c\n\x04seed\x18\x08 \x01(\x03\x12\x10\n\x08isAutoOn\x18\t \x01(\x08\"\x8a\x01\n\x19SpikeSorterCommandRequest\x12\x15\n\rspikeSorterID\x18\x01 \x02(\t\x12\'\n\x03\x63md\x18\x02 \x02(\x0e\x32\x1a.allego.SpikeSorterCommand\x12-\n\x06subCmd\x18\x03 \x02(\x0e\x32\x1d.allego.SpikeSorterSubCommand\"\xcf\x01\n\x10SpikeSorterState\x12+\n\x03sys\x18\x01 \x02(\x0e\x32\x1e.allego.SpikeSorterStateSystem\x12\x14\n\x0c\x66racComplete\x18\x02 \x01(\x01\x12\x0b\n\x03msg\x18\x03 \x01(\t\x12\x12\n\nlaunchTime\x18\x04 \x01(\t\x12\x16\n\x0einitializeTime\x18\x05 \x01(\t\x12\x18\n\x10sessionStartTime\x18\x06 \x01(\t\x12\x17\n\x0fsessionStopTime\x18\x07 \x01(\t\x12\x0c\n\x04isOn\x18\x08 \x01(\x08\"\x81\x01\n\x0cSpkSortIOreq\x12#\n\x04type\x18\x01 \x02(\x0e\x32\x15.allego.SpkSortIOtype\x12\'\n\x06target\x18\x02 \x02(\x0e\x32\x17.allego.SpkSortIOtarget\x12#\n\x04mode\x18\x03 \x02(\x0e\x32\x15.allego.SpkSortIOmode\"i\n\x16SpikeSorterLaunchReply\x12\x15\n\rspikeSorterID\x18\x01 \x02(\t\x12+\n\nsorterType\x18\x02 \x02(\x0e\x32\x17.allego.SpikeSorterType\x12\x0b\n\x03msg\x18\x03 \x01(\t\"0\n\x16GetSpikeSorterIDsReply\x12\x16\n\x0espikeSorterIDs\x18\x01 \x03(\t\"\xa1\x01\n\x1fSpikeSorterGetRasterDataRequest\x12\x15\n\rspikeSorterID\x18\x01 \x02(\t\x12\x12\n\nntvChanIdx\x18\x02 \x03(\x05\x12\x12\n\ntimeWindow\x18\x03 \x02(\x01\x12\x17\n\x0fplotWidthPoints\x18\x04 \x02(\x01\x12\x13\n\x0b\x63omponentID\x18\x05 \x02(\t\x12\x11\n\ttimeRange\x18\x06 \x03(\x01\"\xaa\x01\n\x1aSpikeSorterRasterDataReply\x12<\n\x18spikeTimestampsByChannel\x18\x02 \x03(\x0b\x32\x1a.allego.SpikeTimestampData\x12\x11\n\ttimeRange\x18\x03 \x03(\x01\x12\x16\n\x0etimeStampRange\x18\x04 \x03(\x03\x12\x10\n\x08GPIOData\x18\x05 \x02(\x0c\x12\x11\n\tGPIOShape\x18\x06 \x03(\x05\"3\n\x1aSpikeSorterStandardRequest\x12\x15\n\rspikeSorterID\x18\x01 \x02(\t\"\x86\x03\n\x12SpikeSorterDynamic\x12\x11\n\tisEnabled\x18\x01 \x01(\x08\x12\x43\n\x0fupdatePeriodSec\x18\x02 \x01(\x0b\x32*.allego.SpikeSorterDynamic.UpdatePeriodSec\x12\x35\n\x08\x63riteria\x18\x03 \x01(\x0b\x32#.allego.SpikeSorterDynamic.Criteria\x1a\xa0\x01\n\x08\x43riteria\x12\x31\n\x06\x64\x65tect\x18\x01 \x01(\x0e\x32!.allego.SpikeSorterCriterionLevel\x12\x30\n\x05train\x18\x02 \x01(\x0e\x32!.allego.SpikeSorterCriterionLevel\x12/\n\x04sort\x18\x03 \x01(\x0e\x32!.allego.SpikeSorterCriterionLevel\x1a>\n\x0fUpdatePeriodSec\x12\x0e\n\x06\x64\x65tect\x18\x01 \x01(\x01\x12\r\n\x05train\x18\x02 \x01(\x01\x12\x0c\n\x04sort\x18\x03 \x01(\x01\"\xc5\x01\n\x1cSpikeSorterSetDynamicRequest\x12\x15\n\rspikeSorterID\x18\x01 \x02(\t\x12\x42\n\x07variant\x18\x02 \x03(\x0b\x32\x31.allego.SpikeSorterSetDynamicRequest.VariantEntry\x1aJ\n\x0cVariantEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12)\n\x05value\x18\x02 \x01(\x0b\x32\x1a.allego.SpikeSorterDynamic:\x02\x38\x01\"\xd8\x01\n\x1aSpikeSorterGetDynamicReply\x12\x15\n\rspikeSorterID\x18\x01 \x02(\t\x12\x15\n\rrootVariantID\x18\x02 \x02(\t\x12@\n\x07variant\x18\x03 \x03(\x0b\x32/.allego.SpikeSorterGetDynamicReply.VariantEntry\x1aJ\n\x0cVariantEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12)\n\x05value\x18\x02 \x01(\x0b\x32\x1a.allego.SpikeSorterDynamic:\x02\x38\x01\"l\n\x1fGetSpikeSorterParamCommandReply\x12\x15\n\rspikeSorterID\x18\x01 \x02(\t\x12\x32\n\x03rec\x18\x02 \x03(\x0b\x32%.allego.SpikeSorterParamCommandRecord\"3\n\x12\x43lusterMapOutliers\x12\x0c\n\x04mild\x18\x01 \x03(\x01\x12\x0f\n\x07\x65xtreme\x18\x02 \x03(\x01\"9\n\x13\x43lusterMapQuartiles\x12\n\n\x02q1\x18\x01 \x02(\x01\x12\n\n\x02q2\x18\x02 \x02(\x01\x12\n\n\x02q3\x18\x03 \x02(\x01\"\x83\x01\n\x0f\x43lusterMapStats\x12\x16\n\x0e\x63\x65ntroidCenter\x18\x01 \x03(\x01\x12\x12\n\ncentroidSd\x18\x02 \x03(\x01\x12\x16\n\x0e\x63\x65ntroidSdNorm\x18\x03 \x02(\x01\x12\x19\n\x11intraCentroidDist\x18\x04 \x03(\x01\x12\x11\n\tnumSpikes\x18\x05 \x02(\x03\"f\n\x10\x43lusterMapRecord\x12\r\n\x05label\x18\x01 \x02(\x05\x12&\n\x05stats\x18\x03 \x02(\x0b\x32\x17.allego.ClusterMapStats\x12\x1b\n\x13numSpikesTrainCache\x18\x04 \x02(\x03\"\xa8\x01\n\x13SpikeSortSiteStatus\x12\x12\n\nntvChanIdx\x18\x01 \x02(\x05\x12\x15\n\rorderedLabels\x18\x02 \x03(\x05\x12)\n\x07\x63luster\x18\x03 \x03(\x0b\x32\x18.allego.ClusterMapRecord\x12\x11\n\tnumSpikes\x18\x04 \x03(\x03\x12\x0c\n\x04zeta\x18\x05 \x03(\x01\x12\x1a\n\x12\x61ggregateSpikeRate\x18\x06 \x02(\x01\"\x8f\x01\n\x14SpikeSortPhaseStatus\x12\x11\n\ttimeRange\x18\x01 \x03(\x01\x12\x12\n\nchunkIndex\x18\x02 \x02(\x03\x12\x14\n\x0cupdatePeriod\x18\x03 \x02(\x01\x12\x0c\n\x04\x62\x65ta\x18\x04 \x02(\x01\x12\x12\n\nisComplete\x18\x05 \x02(\x08\x12\x18\n\x10\x66ractionComplete\x18\x06 \x02(\x01\"*\n\x13SpikeSortNbrIdxList\x12\x13\n\x0bntvChanIdxs\x18\x01 \x03(\x05\"\xf2\x02\n\x16SpikeSorterStatusReply\x12\'\n\x05state\x18\x01 \x02(\x0b\x32\x18.allego.SpikeSorterState\x12\x16\n\x0e\x62iointerfaceID\x18\x03 \x02(\t\x12\x14\n\x0c\x64\x61tasourceID\x18\x04 \x02(\t\x12\x10\n\x08sampFreq\x18\x05 \x02(\x01\x12\x19\n\x11\x65nabledNtvChanIdx\x18\x06 \x03(\x05\x12+\n\x05phase\x18\x07 \x03(\x0b\x32\x1c.allego.SpikeSortPhaseStatus\x12\x30\n\x0b\x65nabledSite\x18\x08 \x03(\x0b\x32\x1b.allego.SpikeSortSiteStatus\x12\x12\n\nprobeYield\x18\t \x02(\x01\x12\x11\n\tsortStats\x18\n \x03(\x01\x12\x1b\n\x13\x64\x61tasourceTimeRange\x18\x0c \x03(\x01\x12\x1d\n\x15\x62iointerfaceTimeRange\x18\r \x03(\x01\x12\x12\n\nnumNeurons\x18\x0f \x02(\x03\"R\n\x1eSpikeSorterStatusVariantsReply\x12\x30\n\x08variants\x18\x01 \x03(\x0b\x32\x1e.allego.SpikeSorterStatusReply\"\xfd\x02\n\x19SpikeSorterClassifierSpec\x12\r\n\x05\x61lpha\x18\x01 \x02(\x01\x12\x11\n\talphaBias\x18\x02 \x02(\x01\x12\x16\n\x0e\x64imOutputLayer\x18\x03 \x02(\x05\x12\x16\n\x0e\x65rrorTolerance\x18\x04 \x02(\x01\x12\x19\n\x11hiddenLayerFactor\x18\x05 \x02(\x01\x12\n\n\x02iD\x18\x06 \x02(\t\x12\x16\n\x0einitBiasStdDev\x18\x07 \x02(\x01\x12\x18\n\x10initWeightStdDev\x18\x08 \x02(\x01\x12\x11\n\tisMasking\x18\t \x02(\x08\x12\x10\n\x08l1lambda\x18\n \x02(\x01\x12\x15\n\rminibatchSize\x18\x0b \x02(\x05\x12\x15\n\rmaxNumSamples\x18\x0c \x02(\x05\x12\x11\n\tnumEpochs\x18\r \x02(\x05\x12\x19\n\x11numSamplesTestMSE\x18\x0e \x02(\x05\x12\x1b\n\x13numConcurrentModels\x18\x0f \x02(\x05\x12\x17\n\x0flabelNoiseLevel\x18\x10 \x02(\x01\"u\n\x16SpikeSorterVariantSpec\x12\n\n\x02iD\x18\x01 \x02(\t\x12\x14\n\x0csiteConfigID\x18\x02 \x02(\t\x12\x39\n\x0e\x63lassifierSpec\x18\x03 \x02(\x0b\x32!.allego.SpikeSorterClassifierSpec\"\xcc\x01\n\x0fSpikeSorterSpec\x12\x15\n\rspikeSorterID\x18\x01 \x02(\t\x12\x14\n\x0c\x64\x61tasourceID\x18\x02 \x02(\t\x12\x16\n\x0e\x62iointerfaceID\x18\x03 \x02(\t\x12\x1b\n\x13maxNumSiteNeighbors\x18\x04 \x02(\x05\x12\x18\n\x10neighborRadiusUm\x18\x05 \x02(\x01\x12/\n\x07variant\x18\x06 \x03(\x0b\x32\x1e.allego.SpikeSorterVariantSpec\x12\x0c\n\x04seed\x18\x08 \x02(\x03\"\xc2\x01\n\x19SpikeSorterGetConfigReply\x12%\n\x04spec\x18\x01 \x02(\x0b\x32\x17.allego.SpikeSorterSpec\x12\x16\n\x0e\x61IupdatePeriod\x18\x02 \x02(\x01\x12\x19\n\x11orderedVariantIDs\x18\x03 \x03(\t\x12\x15\n\rrootVariantID\x18\x04 \x02(\t\x12\x34\n\x0csorterStatus\x18\x06 \x03(\x0b\x32\x1e.allego.SpikeSorterStatusReply\"\x90\x02\n\x15SpikeSorterVizRequest\x12\x15\n\rspikeSorterID\x18\x01 \x02(\t\x12\x11\n\tvariantID\x18\x02 \x02(\t\x12\x12\n\nntvChanIdx\x18\x03 \x03(\x05\x12\x13\n\x0bspikeLabels\x18\x04 \x03(\x05\x12\x34\n\x07vizType\x18\x05 \x02(\x0e\x32#.allego.SpikeSorterVizRequest.VType\x12\x14\n\x0cisYAutoScale\x18\x06 \x02(\x08\x12\x0c\n\x04yLim\x18\x07 \x03(\x01\x12\x0e\n\x06\x66igDir\x18\x08 \x02(\t\x12\x11\n\tnumSpikes\x18\t \x02(\x05\"\'\n\x05VType\x12\x1e\n\x1aVIZ_SPIKESORT_TRAIN_SPIKES\x10\x00\"\xc2\x05\n\x1cSpikeSorterGetDashboardReply\x12\"\n\x0c\x65nabledPorts\x18\x01 \x03(\x0e\x32\x0c.allego.Port\x12\x37\n\x07general\x18\x02 \x02(\x0b\x32&.allego.SpikeSorterDashboardGeneralRec\x12\x33\n\tsiteStats\x18\x03 \x02(\x0b\x32 .allego.IndicoDashboardSiteStats\x12\x46\n\tportStats\x18\x04 \x03(\x0b\x32\x33.allego.SpikeSorterGetDashboardReply.PortStatsEntry\x12\x46\n\tsitePanel\x18\x05 \x03(\x0b\x32\x33.allego.SpikeSorterGetDashboardReply.SitePanelEntry\x12J\n\x0bneuronPanel\x18\x06 \x03(\x0b\x32\x35.allego.SpikeSorterGetDashboardReply.NeuronPanelEntry\x12*\n\x02\x61I\x18\x07 \x01(\x0b\x32\x1e.allego.SpikeSorterDashboardAI\x1aR\n\x0ePortStatsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12/\n\x05value\x18\x02 \x01(\x0b\x32 .allego.IndicoDashboardPortStats:\x02\x38\x01\x1aW\n\x0eSitePanelEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x34\n\x05value\x18\x02 \x01(\x0b\x32%.allego.IndicoDashboardSiteIndicators:\x02\x38\x01\x1a[\n\x10NeuronPanelEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x36\n\x05value\x18\x02 \x01(\x0b\x32\'.allego.IndicoDashboardNeuronIndicators:\x02\x38\x01\"\xac\x03\n\x1eSpikeSorterDashboardGeneralRec\x12\'\n\x05state\x18\x01 \x02(\x0b\x32\x18.allego.SpikeSorterState\x12+\n\nsorterType\x18\x02 \x02(\x0e\x32\x17.allego.SpikeSorterType\x12\x10\n\x08sorterID\x18\x03 \x02(\t\x12\x14\n\x0c\x64\x61tasourceID\x18\x04 \x02(\t\x12\x11\n\ttimeRange\x18\x05 \x03(\x01\x12\x15\n\rnumTotalSites\x18\x06 \x02(\x03\x12\x17\n\x0fnumEnabledSites\x18\x07 \x02(\x03\x12\x16\n\x0enumActiveSites\x18\x08 \x02(\x03\x12\x12\n\nnumNeurons\x18\t \x02(\x03\x12\x12\n\nprobeYield\x18\n \x02(\x01\x12\x11\n\tsiteYield\x18\x0b \x02(\x01\x12\x1a\n\x12numSpikesProcessed\x18\x0c \x02(\x03\x12\x18\n\x10numSpikesLabeled\x18\r \x02(\x03\x12\x16\n\x0esortEfficiency\x18\x0e \x02(\x01\x12(\n\x08sinkDesc\x18\x0f \x02(\x0b\x32\x16.allego.FileDescriptor\"\xd5\x01\n\x16SpikeSorterDashboardAI\x12\x0c\n\x04\x62\x65ta\x18\x01 \x02(\x01\x12\"\n\x1a\x65stimatedReaderCompleteSec\x18\x02 \x02(\x01\x12\x1a\n\x12interSessionDurSec\x18\x03 \x02(\x01\x12\x19\n\x11interSessionStart\x18\x04 \x02(\t\x12\x15\n\rsessionDurSec\x18\x05 \x02(\x01\x12\x14\n\x0csessionStart\x18\x06 \x02(\t\x12\x12\n\nsessionIdx\x18\x07 \x02(\x03\x12\x11\n\tstartTime\x18\x08 \x02(\t\"_\n\x18IndicoDashboardSiteStats\x12\x0b\n\x03snr\x18\x01 \x03(\x01\x12\r\n\x05noise\x18\x02 \x03(\x01\x12\x13\n\x0bneuronYield\x18\x03 \x03(\x01\x12\x12\n\nnumNeurons\x18\x04 \x02(\x03\"\x90\x01\n\x18IndicoDashboardPortStats\x12\x0b\n\x03snr\x18\x01 \x03(\x01\x12\r\n\x05noise\x18\x02 \x03(\x01\x12\x13\n\x0bneuronYield\x18\x03 \x03(\x01\x12\x16\n\x0esortEfficiency\x18\x04 \x02(\x01\x12\x17\n\x0fnumEnabledSites\x18\x05 \x02(\x03\x12\x12\n\nnumNeurons\x18\x06 \x02(\x03\"\xd9\x01\n\x1dIndicoDashboardSiteIndicators\x12\x12\n\nntvChanIdx\x18\x01 \x03(\x05\x12\x0b\n\x03snr\x18\x02 \x03(\x01\x12\r\n\x05noise\x18\x03 \x03(\x01\x12\x13\n\x0bneuronYield\x18\x04 \x03(\x01\x12\x16\n\x0esortEfficiency\x18\x05 \x03(\x01\x12\x11\n\tspikeRate\x18\x06 \x03(\x01\x12\x11\n\tposProbeX\x18\x07 \x03(\x01\x12\x11\n\tposProbeY\x18\x08 \x03(\x01\x12\x11\n\tposProbeZ\x18\t \x03(\x01\x12\x0f\n\x07siteNum\x18\n \x03(\x05\"\xb1\x01\n\x1fIndicoDashboardNeuronIndicators\x12\x10\n\x08neuronID\x18\x01 \x03(\t\x12\x0b\n\x03snr\x18\x02 \x03(\x01\x12\x11\n\tspikeRate\x18\x03 \x03(\x01\x12\x11\n\tposProbeX\x18\x04 \x03(\x01\x12\x11\n\tposProbeY\x18\x05 \x03(\x01\x12\x11\n\tposProbeZ\x18\x06 \x03(\x01\x12\x0f\n\x07siteNum\x18\x07 \x03(\x05\x12\x12\n\nspikeLabel\x18\x08 \x03(\x05\"\xa7\x02\n\x1cSpikeSorterDashboardSiteDesc\x12\x1a\n\x04port\x18\x01 \x02(\x0e\x32\x0c.allego.Port\x12\x12\n\nntvChanIdx\x18\x02 \x02(\x05\x12\x10\n\x08posProbe\x18\x03 \x03(\x01\x12\x10\n\x08posBrain\x18\x04 \x03(\x01\x12\x0e\n\x06snrMax\x18\x05 \x02(\x01\x12\x0e\n\x06snrMin\x18\x06 \x02(\x01\x12\x12\n\nnoiseLevel\x18\x07 \x02(\x01\x12\x12\n\nnumNeurons\x18\x08 \x02(\x03\x12/\n\rindicoCluster\x18\n \x03(\x0b\x32\x18.allego.ClusterMapRecord\x12\x1c\n\x14pileWaveformMinValue\x18\r \x02(\x02\x12\x1c\n\x14pileWaveformMaxValue\x18\x0e \x02(\x02\"\xd2\x01\n\x1eSpikeSorterDashboardNeuronDesc\x12\x1a\n\x04port\x18\x01 \x02(\x0e\x32\x0c.allego.Port\x12&\n\x04\x64\x65sc\x18\x02 \x02(\x0b\x32\x18.allego.NeuronDescriptor\x12\x0b\n\x03snr\x18\x05 \x02(\x01\x12\x0f\n\x07IsihMax\x18\x06 \x02(\x01\x12\x15\n\rIsihArgmaxSec\x18\x07 \x02(\x01\x12\x15\n\rIsihArgmaxIdx\x18\x08 \x02(\x03\x12\x10\n\x08IsihMean\x18\t \x02(\x01\x12\x0e\n\x06IsihSd\x18\n \x02(\x01\"\xfa\x01\n SpikeSorterDashboardNeuronDetail\x12&\n\x04\x64\x65sc\x18\x01 \x02(\x0b\x32\x18.allego.NeuronDescriptor\x12\x18\n\x10templateWaveform\x18\x02 \x03(\x02\x12\x1a\n\x12templateSdWaveform\x18\x03 \x03(\x02\x12\x18\n\x10pileWaveformData\x18\x04 \x03(\x02\x12\x1e\n\x16numSamplesPileWaveform\x18\x05 \x02(\x05\x12\x18\n\x10numPileWaveforms\x18\x06 \x02(\x05\x12$\n\x04isih\x18\x0b \x02(\x0b\x32\x16.allego.NeuronHistData\"\xa9\x01\n\x1eSpikeSorterDashboardSiteDetail\x12\x32\n\x04\x64\x65sc\x18\x01 \x02(\x0b\x32$.allego.SpikeSorterDashboardSiteDesc\x12\x19\n\x11orderedSpikeLabel\x18\x02 \x03(\x05\x12\x38\n\x06neuron\x18\x03 \x03(\x0b\x32(.allego.SpikeSorterDashboardNeuronDetail*\x80\x01\n\x0fSpikeSorterType\x12\x0f\n\x0bSORTER_NULL\x10\x00\x12\x15\n\x11SORTER_VEX_STREAM\x10\x01\x12\x18\n\x14SORTER_SPIKES_STREAM\x10\x02\x12\x13\n\x0fSORTER_VEX_FILE\x10\x03\x12\x16\n\x12SORTER_SPIKES_FILE\x10\x04*\xb0\x01\n\x12SpikeSorterCommand\x12\x11\n\rSORTER_CMD_ON\x10\x00\x12\x12\n\x0eSORTER_CMD_OFF\x10\x01\x12\x19\n\x15SORTER_CMD_CLEAR_SORT\x10\x02\x12\x15\n\x11SORTER_CMD_REBASE\x10\x04\x12\x13\n\x0fSORTER_CMD_INIT\x10\x07\x12\x17\n\x13SORTER_CMD_LOCALIZE\x10\x08\x12\x13\n\x0fSORTER_CMD_SORT\x10\t*/\n\x15SpikeSorterSubCommand\x12\x16\n\x12SORTER_SUBCMD_NULL\x10\x00*@\n\x0fSpikeSorterMode\x12\x16\n\x12SORTER_MODE_STREAM\x10\x00\x12\x15\n\x11SORTER_MODE_BATCH\x10\x01*I\n\x16SpikeSorterStateSystem\x12\n\n\x06SYS_ON\x10\x00\x12\x0b\n\x07SYS_OFF\x10\x01\x12\x16\n\x12SYS_NOT_CONFIGURED\x10\x02*f\n\rSpkSortIOtype\x12\x17\n\x13SPK_SORT_IO_NETWORK\x10\x00\x12\x17\n\x13SPK_SORT_IO_TRAINER\x10\x01\x12#\n\x1fSPK_SORT_IO_NETWORK_AND_TRAINER\x10\x02*>\n\x0fSpkSortIOtarget\x12\x14\n\x10SPK_SORT_IO_FSYS\x10\x00\x12\x15\n\x11SPK_SORT_IO_CLOUD\x10\x01*;\n\rSpkSortIOmode\x12\x14\n\x10SPK_SORT_IO_SAVE\x10\x00\x12\x14\n\x10SPK_SORT_IO_LOAD\x10\x01*u\n\x19SpikeSorterCriterionLevel\x12\x10\n\x0cSORTER_LEAST\x10\x00\x12\x0f\n\x0bSORTER_LESS\x10\x01\x12\x13\n\x0fSORTER_BALANCED\x10\x02\x12\x0f\n\x0bSORTER_MORE\x10\x03\x12\x0f\n\x0bSORTER_MOST\x10\x04*\x84\x01\n\x07SS_STAT\x12\x08\n\x04MEAN\x10\x00\x12\x06\n\x02SD\x10\x01\x12\x08\n\x04MODE\x10\x02\x12\x07\n\x03MIN\x10\x03\x12\x07\n\x03MAX\x10\x04\x12\x0e\n\nMODE_COUNT\x10\x05\x12\n\n\x06MEDIAN\x10\x06\x12\x07\n\x03Q25\x10\x07\x12\x07\n\x03Q75\x10\x08\x12\x08\n\x04SKEW\x10\t\x12\x0c\n\x08KURTOSIS\x10\n\x12\x05\n\x01N\x10\x0b\x42\x15Z\x13internal/radix/grpc')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'spikesorter_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z\023internal/radix/grpc'
  _SPIKESORTERSETDYNAMICREQUEST_VARIANTENTRY._options = None
  _SPIKESORTERSETDYNAMICREQUEST_VARIANTENTRY._serialized_options = b'8\001'
  _SPIKESORTERGETDYNAMICREPLY_VARIANTENTRY._options = None
  _SPIKESORTERGETDYNAMICREPLY_VARIANTENTRY._serialized_options = b'8\001'
  _SPIKESORTERGETDASHBOARDREPLY_PORTSTATSENTRY._options = None
  _SPIKESORTERGETDASHBOARDREPLY_PORTSTATSENTRY._serialized_options = b'8\001'
  _SPIKESORTERGETDASHBOARDREPLY_SITEPANELENTRY._options = None
  _SPIKESORTERGETDASHBOARDREPLY_SITEPANELENTRY._serialized_options = b'8\001'
  _SPIKESORTERGETDASHBOARDREPLY_NEURONPANELENTRY._options = None
  _SPIKESORTERGETDASHBOARDREPLY_NEURONPANELENTRY._serialized_options = b'8\001'
  _SPIKESORTERTYPE._serialized_start=7565
  _SPIKESORTERTYPE._serialized_end=7693
  _SPIKESORTERCOMMAND._serialized_start=7696
  _SPIKESORTERCOMMAND._serialized_end=7872
  _SPIKESORTERSUBCOMMAND._serialized_start=7874
  _SPIKESORTERSUBCOMMAND._serialized_end=7921
  _SPIKESORTERMODE._serialized_start=7923
  _SPIKESORTERMODE._serialized_end=7987
  _SPIKESORTERSTATESYSTEM._serialized_start=7989
  _SPIKESORTERSTATESYSTEM._serialized_end=8062
  _SPKSORTIOTYPE._serialized_start=8064
  _SPKSORTIOTYPE._serialized_end=8166
  _SPKSORTIOTARGET._serialized_start=8168
  _SPKSORTIOTARGET._serialized_end=8230
  _SPKSORTIOMODE._serialized_start=8232
  _SPKSORTIOMODE._serialized_end=8291
  _SPIKESORTERCRITERIONLEVEL._serialized_start=8293
  _SPIKESORTERCRITERIONLEVEL._serialized_end=8410
  _SS_STAT._serialized_start=8413
  _SS_STAT._serialized_end=8545
  _SPIKESORTERLAUNCHREQUEST._serialized_start=64
  _SPIKESORTERLAUNCHREQUEST._serialized_end=324
  _SPIKESORTERCOMMANDREQUEST._serialized_start=327
  _SPIKESORTERCOMMANDREQUEST._serialized_end=465
  _SPIKESORTERSTATE._serialized_start=468
  _SPIKESORTERSTATE._serialized_end=675
  _SPKSORTIOREQ._serialized_start=678
  _SPKSORTIOREQ._serialized_end=807
  _SPIKESORTERLAUNCHREPLY._serialized_start=809
  _SPIKESORTERLAUNCHREPLY._serialized_end=914
  _GETSPIKESORTERIDSREPLY._serialized_start=916
  _GETSPIKESORTERIDSREPLY._serialized_end=964
  _SPIKESORTERGETRASTERDATAREQUEST._serialized_start=967
  _SPIKESORTERGETRASTERDATAREQUEST._serialized_end=1128
  _SPIKESORTERRASTERDATAREPLY._serialized_start=1131
  _SPIKESORTERRASTERDATAREPLY._serialized_end=1301
  _SPIKESORTERSTANDARDREQUEST._serialized_start=1303
  _SPIKESORTERSTANDARDREQUEST._serialized_end=1354
  _SPIKESORTERDYNAMIC._serialized_start=1357
  _SPIKESORTERDYNAMIC._serialized_end=1747
  _SPIKESORTERDYNAMIC_CRITERIA._serialized_start=1523
  _SPIKESORTERDYNAMIC_CRITERIA._serialized_end=1683
  _SPIKESORTERDYNAMIC_UPDATEPERIODSEC._serialized_start=1685
  _SPIKESORTERDYNAMIC_UPDATEPERIODSEC._serialized_end=1747
  _SPIKESORTERSETDYNAMICREQUEST._serialized_start=1750
  _SPIKESORTERSETDYNAMICREQUEST._serialized_end=1947
  _SPIKESORTERSETDYNAMICREQUEST_VARIANTENTRY._serialized_start=1873
  _SPIKESORTERSETDYNAMICREQUEST_VARIANTENTRY._serialized_end=1947
  _SPIKESORTERGETDYNAMICREPLY._serialized_start=1950
  _SPIKESORTERGETDYNAMICREPLY._serialized_end=2166
  _SPIKESORTERGETDYNAMICREPLY_VARIANTENTRY._serialized_start=1873
  _SPIKESORTERGETDYNAMICREPLY_VARIANTENTRY._serialized_end=1947
  _GETSPIKESORTERPARAMCOMMANDREPLY._serialized_start=2168
  _GETSPIKESORTERPARAMCOMMANDREPLY._serialized_end=2276
  _CLUSTERMAPOUTLIERS._serialized_start=2278
  _CLUSTERMAPOUTLIERS._serialized_end=2329
  _CLUSTERMAPQUARTILES._serialized_start=2331
  _CLUSTERMAPQUARTILES._serialized_end=2388
  _CLUSTERMAPSTATS._serialized_start=2391
  _CLUSTERMAPSTATS._serialized_end=2522
  _CLUSTERMAPRECORD._serialized_start=2524
  _CLUSTERMAPRECORD._serialized_end=2626
  _SPIKESORTSITESTATUS._serialized_start=2629
  _SPIKESORTSITESTATUS._serialized_end=2797
  _SPIKESORTPHASESTATUS._serialized_start=2800
  _SPIKESORTPHASESTATUS._serialized_end=2943
  _SPIKESORTNBRIDXLIST._serialized_start=2945
  _SPIKESORTNBRIDXLIST._serialized_end=2987
  _SPIKESORTERSTATUSREPLY._serialized_start=2990
  _SPIKESORTERSTATUSREPLY._serialized_end=3360
  _SPIKESORTERSTATUSVARIANTSREPLY._serialized_start=3362
  _SPIKESORTERSTATUSVARIANTSREPLY._serialized_end=3444
  _SPIKESORTERCLASSIFIERSPEC._serialized_start=3447
  _SPIKESORTERCLASSIFIERSPEC._serialized_end=3828
  _SPIKESORTERVARIANTSPEC._serialized_start=3830
  _SPIKESORTERVARIANTSPEC._serialized_end=3947
  _SPIKESORTERSPEC._serialized_start=3950
  _SPIKESORTERSPEC._serialized_end=4154
  _SPIKESORTERGETCONFIGREPLY._serialized_start=4157
  _SPIKESORTERGETCONFIGREPLY._serialized_end=4351
  _SPIKESORTERVIZREQUEST._serialized_start=4354
  _SPIKESORTERVIZREQUEST._serialized_end=4626
  _SPIKESORTERVIZREQUEST_VTYPE._serialized_start=4587
  _SPIKESORTERVIZREQUEST_VTYPE._serialized_end=4626
  _SPIKESORTERGETDASHBOARDREPLY._serialized_start=4629
  _SPIKESORTERGETDASHBOARDREPLY._serialized_end=5335
  _SPIKESORTERGETDASHBOARDREPLY_PORTSTATSENTRY._serialized_start=5071
  _SPIKESORTERGETDASHBOARDREPLY_PORTSTATSENTRY._serialized_end=5153
  _SPIKESORTERGETDASHBOARDREPLY_SITEPANELENTRY._serialized_start=5155
  _SPIKESORTERGETDASHBOARDREPLY_SITEPANELENTRY._serialized_end=5242
  _SPIKESORTERGETDASHBOARDREPLY_NEURONPANELENTRY._serialized_start=5244
  _SPIKESORTERGETDASHBOARDREPLY_NEURONPANELENTRY._serialized_end=5335
  _SPIKESORTERDASHBOARDGENERALREC._serialized_start=5338
  _SPIKESORTERDASHBOARDGENERALREC._serialized_end=5766
  _SPIKESORTERDASHBOARDAI._serialized_start=5769
  _SPIKESORTERDASHBOARDAI._serialized_end=5982
  _INDICODASHBOARDSITESTATS._serialized_start=5984
  _INDICODASHBOARDSITESTATS._serialized_end=6079
  _INDICODASHBOARDPORTSTATS._serialized_start=6082
  _INDICODASHBOARDPORTSTATS._serialized_end=6226
  _INDICODASHBOARDSITEINDICATORS._serialized_start=6229
  _INDICODASHBOARDSITEINDICATORS._serialized_end=6446
  _INDICODASHBOARDNEURONINDICATORS._serialized_start=6449
  _INDICODASHBOARDNEURONINDICATORS._serialized_end=6626
  _SPIKESORTERDASHBOARDSITEDESC._serialized_start=6629
  _SPIKESORTERDASHBOARDSITEDESC._serialized_end=6924
  _SPIKESORTERDASHBOARDNEURONDESC._serialized_start=6927
  _SPIKESORTERDASHBOARDNEURONDESC._serialized_end=7137
  _SPIKESORTERDASHBOARDNEURONDETAIL._serialized_start=7140
  _SPIKESORTERDASHBOARDNEURONDETAIL._serialized_end=7390
  _SPIKESORTERDASHBOARDSITEDETAIL._serialized_start=7393
  _SPIKESORTERDASHBOARDSITEDETAIL._serialized_end=7562
# @@protoc_insertion_point(module_scope)



import grpc
import numpy as np
import pandas as pd
import warnings
from radiens.exceptions.grpc_error import handle_grpc_error
from radiens.grpc import (allegoserver_pb2, allegoserver_pb2_grpc,
                          common_pb2, spikesorter_pb2)
from radiens.signal_group.SignalGroup import SignalGroup
from radiens.utils.config import new_server_channel
from radiens.utils.constants import (DOUT_MODE_OUT, PORT_, PRIMARY_CACHE_STREAM_GROUP_ID, PORT_ENUM,
                                     HEADSTAGE_ALIAS, PROBE_ALIAS, SYSTEM_MODE, SYSTEM_MODE_DECODE,
                                     APPS_, SPK_SORTER_, SPK_SORTER_STATE_DECODE)

CLIENT_NAME = 'Allego'

# ====== life cycle ======


def restart(addr, mode: str):
    if mode not in SYSTEM_MODE.keys():
        raise ValueError('invalid system type {}'.format(mode))

    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            stub.Restart(allegoserver_pb2.RestartRequest(mode=SYSTEM_MODE[mode]))
        except grpc.RpcError as ex:
            handle_grpc_error(ex, CLIENT_NAME)

# ====== workspace ======


def workspace_save(addr, is_force_overwrite=True, tags='', notes=''):
    annotate = common_pb2.AnnotateBundle(tags=tags, notes=notes)
    req = common_pb2.WorkspaceControlRequest(cmd=common_pb2.WSPACE_Save,
                                             annotation=annotate, isForceOverwrite=is_force_overwrite)
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            stub.WorkspaceControl(req)
        except grpc.RpcError as ex:
            handle_grpc_error(ex, CLIENT_NAME)


def workspace_save_as(addr, wspace_id: str, is_force_overwrite=True, tags='', notes=''):
    annotate = common_pb2.AnnotateBundle(tags=tags, notes=notes)
    req = common_pb2.WorkspaceControlRequest(cmd=common_pb2.WSPACE_Save,
                                             workspaceID=wspace_id,
                                             annotation=annotate, isForceOverwrite=is_force_overwrite)
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            stub.WorkspaceControl(req)
        except grpc.RpcError as ex:
            handle_grpc_error(ex, CLIENT_NAME)


def workspace_switch(addr, wspace_id: str):
    req = common_pb2.WorkspaceControlRequest(cmd=common_pb2.WSPACE_Switch,
                                             workspaceID=wspace_id)
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            stub.WorkspaceControl(req)
        except grpc.RpcError as ex:
            handle_grpc_error(ex, CLIENT_NAME)


def workspace_delete(addr, wspace_id: str):
    req = common_pb2.WorkspaceControlRequest(cmd=common_pb2.WSPACE_Delete,
                                             workspaceID=wspace_id)
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            stub.WorkspaceControl(req)
        except grpc.RpcError as ex:
            handle_grpc_error(ex, CLIENT_NAME)


def workspace_current(addr):
    req = common_pb2.GetWorkspaceRequest(cmd=common_pb2.GET_WSPACE_Current, appMask=common_pb2.Allego)
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            resp = stub.GetWorkspace(req)
        except grpc.RpcError as ex:
            handle_grpc_error(ex, CLIENT_NAME)
        d = {'ID': [], 'app': [], 'last_used': [], 'last_modified': [], 'is_modified': [], 'notes': [], 'tags': []}
        for _, v in resp.workspaceDesc.items():
            d['ID'].append(v.iD)
            d['app'].append(APPS_[v.app])
            d['last_used'].append(v.timestampLastUsed)
            d['last_modified'].append(v.timestampModified)
            d['is_modified'].append(v.isModified)
            d['tags'].append(v.annotation.tags)
            d['notes'].append(v.annotation.notes)
        return pd.DataFrame(d)


def workspace_list(addr):
    req = common_pb2.GetWorkspaceRequest(cmd=common_pb2.GET_WSPACE_List, appMask=common_pb2.Allego)
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            resp = stub.GetWorkspace(req)
        except grpc.RpcError as ex:
            handle_grpc_error(ex, CLIENT_NAME)
        d = {'ID': [], 'app': [], 'last_used': [], 'last_modified': [], 'is_modified': [], 'notes': [], 'tags': []}
        for _, v in resp.workspaceDesc.items():
            d['ID'].append(v.iD)
            d['app'].append(APPS_[v.app])
            d['last_used'].append(v.timestampLastUsed)
            d['last_modified'].append(v.timestampModified)
            d['is_modified'].append(v.isModified)
            d['tags'].append(v.annotation.tags)
            d['notes'].append(v.annotation.notes)
        return pd.DataFrame(d)

# ====== getters ======


def get_config(addr):
    cfg = {}
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            res = stub.GetConfig(common_pb2.StandardRequest())
        except grpc.RpcError as ex:
            handle_grpc_error(ex, CLIENT_NAME)
        else:
            cable_dict = {}
            for p in ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L']:
                try:
                    cable_dict[p] = eval('res.cableLengths.{}'.format(p))
                except AttributeError:
                    pass
            cfg = {'system_server_port': res.allegoCoreServerPort,
                   'base_samp_freq': res.baseSampFreq,
                   'stream_loop_dur_ms': res.streamLoopDurMs,
                   'headstage_cable_lens': cable_dict,
                   'backbone_mode': SYSTEM_MODE_DECODE[res.backboneMode]}

        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            res = stub.GetConfigRecording(common_pb2.StandardRequest())
        except grpc.RpcError as ex:
            handle_grpc_error(ex, CLIENT_NAME)
        else:
            cfg['base_filename'] = res.baseFileName
            cfg['filepath'] = res.baseFilePath

        return cfg


def get_ports(addr):
    req = common_pb2.SignalGroupIDRequest()
    req.streamGroupId = PRIMARY_CACHE_STREAM_GROUP_ID
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            res = stub.ListSensorSpecs(req)
        except grpc.RpcError as ex:
            handle_grpc_error(ex, CLIENT_NAME)
        ports = {}
        for _, v in enumerate(res.ports):
            ports[PORT_ENUM[v.port]] = v.channelCount
        return ports


def get_sensors(addr):
    req = common_pb2.SignalGroupIDRequest()
    req.streamGroupId = PRIMARY_CACHE_STREAM_GROUP_ID
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            res = stub.ListSensorSpecs(req)
        except grpc.RpcError as ex:
            handle_grpc_error(ex, CLIENT_NAME)
        hstages = {}
        for _, v in enumerate(res.headstages):
            hstages[v.name.strip('hstg__')] = v.channelCount
        probes = {}
        for _, v in enumerate(res.probes):
            probes[v.name.strip('probe__')] = v.channelCount
        return {'headstages': hstages, 'probes': probes}


def get_status(addr):
    with new_server_channel(addr) as chan:
        AllegoCoreStub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            res = AllegoCoreStub.GetStatus(common_pb2.StandardRequest())
        except grpc.RpcError as ex:
            handle_grpc_error(ex, CLIENT_NAME)
        else:
            return {'streaming': {'mode': 'on' if res.streaming.streamMode == 1 else 'off',
                                  'time_range': [res.streaming.primaryCacheTRange[0], res.streaming.primaryCacheTRange[1]],
                                  'hardware_memory': res.streaming.hardwareMemoryLevel},
                    'recording': {'mode': 'on' if res.recording.recordMode == 1 else 'off',
                                  'active_filename': res.recording.activeFileName,
                                  'duration': res.recording.duration,
                                  'error': res.recording.error}}


def get_signal_group(addr, stream_group_id=None):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            req = common_pb2.SignalGroupIDRequest()
            req.streamGroupId = stream_group_id

            raw = stub.GetSignalGroup(req)
        except grpc.RpcError as ex:
            handle_grpc_error(ex, CLIENT_NAME)
        else:
            return SignalGroup(raw)


def get_signals(addr, stream_group_id: str, samp_freq: float):
    sigarray = []
    time_range = []
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.Pcache1Stub(chan)
        req = common_pb2.GetSignalsRequest(streamGroupId=stream_group_id)
        req.params.timeWindow = 1
        req.params.spacing = 0
        req.params.magnitude = -1
        req.params.plotWidthPoints = samp_freq
        req.params.gpioOnTop = False
        req.params.auxMagnitude = -1
        req.params.componentID = 'allego_python_client'

        try:
            raw = stub.GetSignals(req)
        except grpc.RpcError as ex:
            handle_grpc_error(ex, CLIENT_NAME)
        else:
            sigarray = np.frombuffer(raw.data, dtype=np.float32)
            sigarray = np.reshape(sigarray, (raw.shape[0], raw.shape[1]))
            time_range = [raw.timeRange[0], raw.timeRange[1]]

            return sigarray, time_range


def get_digital_out_states(addr):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            res = stub.GetDIOReg(common_pb2.StandardRequest())
        except grpc.RpcError as ex:
            handle_grpc_error(ex, CLIENT_NAME)
        else:
            return {
                'digital_outs_mode': DOUT_MODE_OUT[res.mode],
                'states': [{'chan_idx': i.ntvChanIdx, 'state': i.manualState} for i in res.doutChanRegisters]
            }


# ======= setters ========

def set_digital_out_manual(addr, dout1_state: bool, dout2_state: bool):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)

        req1 = allegoserver_pb2.DIOModeManualRequest()
        req1.chanIdx = 0
        req1.state = dout1_state

        req2 = allegoserver_pb2.DIOModeManualRequest()
        req2.chanIdx = 1
        req2.state = dout2_state

        try:
            stub.SetDIOManual(req1)
            stub.SetDIOManual(req2)
        except grpc.RpcError as ex:
            handle_grpc_error(ex, CLIENT_NAME)


def set_fs(addr, fs: int):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            stub.SetConfigCore(allegoserver_pb2.SetConfigCoreRequest(sampFreq=float(fs)))
        except grpc.RpcError as ex:
            handle_grpc_error(ex, CLIENT_NAME)


def set_sensor(addr, port: str, hstage: str, probe: str):
    connected_ports = get_ports(addr)
    if port not in connected_ports.keys():
        raise ValueError('port {} is not valid or is connected'.format(port))
    pb_port = PORT_[port]

    sensors = get_sensors(addr)
    if hstage in HEADSTAGE_ALIAS.keys():
        pb_hstage = 'hstg__{}'.format(HEADSTAGE_ALIAS[hstage])
    elif hstage in sensors['headstages'].keys():
        pb_hstage = 'hstg__{}'.format(hstage)
    else:
        raise ValueError('headstage {} is not an alias nor an available headstage ID'.format(hstage))

    if probe in PROBE_ALIAS.keys():
        pb_probe = 'probe__{}'.format(PROBE_ALIAS[probe])
    elif probe in sensors['probes'].keys():
        pb_probe = 'probe__{}'.format(probe)
    else:
        raise ValueError('probe {} is not an alias nor an available probe ID'.format(hstage))

    port_num_chan = connected_ports[port]
    hstg_num_chan = sensors['headstages'][pb_hstage.strip('hstg__')]
    probe_num_chan = sensors['probes'][pb_probe.strip('probe__')]
    if port_num_chan != hstg_num_chan:
        warnings.warn('port {} number of channels={} does not match headstage number of channels={}'.format(
                      connected_ports[port], port_num_chan, hstg_num_chan))
    if hstg_num_chan != probe_num_chan:
        warnings.warn('headstage {} number of channels={} does not match probe number of channels={}'.format(
                      sensors['headstages'][pb_hstage.strip('hstg__')], hstg_num_chan, probe_num_chan))

    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            stub.SetSensor(common_pb2.SetSensorRequest(streamGroupId=PRIMARY_CACHE_STREAM_GROUP_ID,
                                                       port=pb_port, headstageId=pb_hstage, probeId=pb_probe))
        except grpc.RpcError as ex:
            handle_grpc_error(ex, CLIENT_NAME)


def set_stream_state(addr, state: int):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            stub.SetStreamState(allegoserver_pb2.SetStreamRequest(mode=state))
        except grpc.RpcError as ex:
            handle_grpc_error(ex, CLIENT_NAME)


def set_record_state(addr, state: int):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            stub.SetRecordState(allegoserver_pb2.SetRecordRequest(mode=state))
        except grpc.RpcError as ex:
            handle_grpc_error(ex, CLIENT_NAME)


def set_recording_config(addr, filename: str, filepath: str, index: int, time_stamp: str):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.AllegoCoreStub(chan)
        try:
            stub.SetConfigRecording(allegoserver_pb2.ConfigRecording(baseFileName=filename,
                                                                     baseFilePath=filepath,
                                                                     datasource_idx=index,
                                                                     timeStamp=time_stamp))
        except grpc.RpcError as ex:
            handle_grpc_error(ex, CLIENT_NAME)


def set_time_to_cache_head(addr, stream_group_id=None):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.Pcache1Stub(chan)
        req = common_pb2.SignalGroupIDRequest()
        req.streamGroupId = stream_group_id
        try:
            stub.SetTimeRangeToHead(req)
        except grpc.RpcError as ex:
            handle_grpc_error(ex, CLIENT_NAME)

# ====== spike sorter ======


def sorter_cmd(addr, cmd: str):
    if cmd not in SPK_SORTER_.keys():
        raise ValueError('invalid spike sorter command={}, valid commands are {}'.format(cmd, list(SPK_SORTER_.keys())))

    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.Neurons1Stub(chan)
        try:
            stub.SpikeSorterCommand(spikesorter_pb2.SpikeSorterCommandRequest(
                cmd=SPK_SORTER_[cmd], subCmd=spikesorter_pb2.SORTER_SUBCMD_NULL, spikeSorterID=''))
        except grpc.RpcError as ex:
            handle_grpc_error(ex, CLIENT_NAME)


def sorter_get_params(addr):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.Neurons1Stub(chan)
        try:
            resp = stub.SpikeSorterGetParam(spikesorter_pb2.SpikeSorterStandardRequest(spikeSorterID=''))
        except grpc.RpcError as ex:
            handle_grpc_error(ex, CLIENT_NAME)
        p = {'ntv_chan_idx': [], 'chan_enabled?': [], 'thr_set?': [], 'use_sd?': [], 'thr': [], 'thr_sd': [], 'thr_wdw': [], 'peak_wdw': [],
             'shadow': [], 'weak_thr': [], 'thr_wdw_pts': [], 'peak_wdw_pts': [], 'shadow_pts': []}
        for v in resp.rec:
            p['ntv_chan_idx'].append(v.ntvChanIdx)
            p['chan_enabled?'].append(v.isEnabled)
            p['thr_set?'].append(v.isSetThr)
            p['use_sd?'].append(v.isSD)
            p['thr'].append(np.array(v.thr))
            p['thr_sd'].append(np.array(v.thrSd))
            p['thr_wdw'].append(np.around(np.array(v.thrWdw) * 1000.0, 5))
            p['peak_wdw'].append(np.around(np.array(v.peakWdw) * 1000.0, 5))
            p['shadow'].append(np.around(np.array(v.shadow) * 1000.0, 5))
            p['weak_thr'].append(np.array(v.weakThr))
            p['thr_wdw_pts'].append(v.thrWdwPts)
            p['peak_wdw_pts'].append(v.peakWdwPts)
            p['shadow_pts'].append(v.shadowPts)
        return pd.DataFrame(p)


def sorter_get_state(addr):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.Neurons1Stub(chan)
        try:
            resp = stub.SpikeSorterGetState(spikesorter_pb2.SpikeSorterStandardRequest(spikeSorterID=''))
        except grpc.RpcError as ex:
            handle_grpc_error(ex, CLIENT_NAME)
        return {'sorter': SPK_SORTER_STATE_DECODE[resp.sys], 'msg': resp.msg,
                'time_initialize': resp.initializeTime, 'time_start': resp.sessionStartTime, 'time_stop': resp.sessionStopTime}


def sorter_get_dashboard(addr):
    with new_server_channel(addr) as chan:
        stub = allegoserver_pb2_grpc.Neurons1Stub(chan)
        try:
            resp = stub.SpikeSorterGetDashboard(spikesorter_pb2.SpikeSorterStandardRequest(spikeSorterID=''))
        except grpc.RpcError as ex:
            handle_grpc_error(ex, CLIENT_NAME)
    sum_stats = []
    sum_stats.append(pd.DataFrame({'stat': ['probe-snr', 'probe-noise', 'probe-units'],
                                   'port': ['all', 'all', 'all'],
                                   'N': [resp.siteStats.snr[11], resp.siteStats.noise[11], resp.siteStats.neuronYield[11]],
                                   'mean': [resp.siteStats.snr[0], resp.siteStats.noise[0], resp.siteStats.neuronYield[0]],
                                   'sd': [resp.siteStats.snr[1], resp.siteStats.noise[1], resp.siteStats.neuronYield[1]],
                                   'min': [resp.siteStats.snr[3], resp.siteStats.noise[3], resp.siteStats.neuronYield[3]],
                                   'max': [resp.siteStats.snr[4], resp.siteStats.noise[4], resp.siteStats.neuronYield[4]],
                                   'median': [resp.siteStats.snr[6], resp.siteStats.noise[6], resp.siteStats.neuronYield[6]],
                                   'mode': [np.NaN, np.NaN, resp.siteStats.neuronYield[2]],
                                   'mode_cnt': [np.NaN, np.NaN, resp.siteStats.neuronYield[5]],
                                   'q25': [resp.siteStats.snr[7], resp.siteStats.noise[7], resp.siteStats.neuronYield[7]],
                                   'q75': [resp.siteStats.snr[8], resp.siteStats.noise[8], resp.siteStats.neuronYield[8]],
                                   'skew': [resp.siteStats.snr[9], resp.siteStats.noise[9], resp.siteStats.neuronYield[9]],
                                   'kurtosis': [resp.siteStats.snr[10], resp.siteStats.noise[10], resp.siteStats.neuronYield[10]],
                                   }))
    for k in resp.portStats:
        v = resp.portStats[k]
        sum_stats.append(pd.DataFrame({'stat': ['port-snr', 'port-noise', 'port-units'],
                                       'port': [PORT_ENUM[k], PORT_ENUM[k], PORT_ENUM[k]],
                                       'N': [v.snr[11], v.noise[11], v.neuronYield[11]],
                                       'mean': [v.snr[0], v.noise[0], v.neuronYield[0]],
                                       'sd': [v.snr[1], v.noise[1], v.neuronYield[1]],
                                       'min': [v.snr[3], v.noise[3], v.neuronYield[3]],
                                       'max': [v.snr[4], v.noise[4], v.neuronYield[4]],
                                       'median': [v.snr[6], v.noise[6], v.neuronYield[6]],
                                       'mode': [np.NaN, np.NaN, v.neuronYield[2]],
                                       'mode_cnt': [np.NaN, np.NaN, v.neuronYield[5]],
                                       'q25': [v.snr[7], v.noise[7], v.neuronYield[7]],
                                       'q75': [v.snr[8], v.noise[8], v.neuronYield[8]],
                                       'skew': [v.snr[9], v.noise[9], v.neuronYield[9]],
                                       'kurtosis': [v.snr[10], v.noise[10], v.neuronYield[10]],
                                       }))
    df_sum_stats = pd.concat(sum_stats, axis=0)
    return {'enabled_ports': resp.enabledPorts, 'general': {'sorter': SPK_SORTER_STATE_DECODE[resp.general.state.sys],
                                                            'msg': resp.general.state.msg,
                                                            'time_range': np.array(resp.general.timeRange),
                                                            'num_sites_total': resp.general.numTotalSites,
                                                            'num_sites_enabled': resp.general.numEnabledSites,
                                                            'num_sites_active': resp.general.numActiveSites,
                                                            'num_units': resp.general.numNeurons,
                                                            'probe_yield_mean': resp.general.probeYield,
                                                            'site_yield_mean': resp.general.siteYield,
                                                            'num_spikes_processed': resp.general.numSpikesProcessed,
                                                            'num_spikes_labeled': resp.general.numSpikesLabeled,
                                                            'sort_efficiency': resp.general.sortEfficiency,
                                                            'spikes_file_path': resp.general.sinkDesc.path,
                                                            'spikes_file_base_name': resp.general.sinkDesc.baseName,
                                                            },
            'summary_stats': df_sum_stats}



import radiens.api.api_allego as api
from radiens.utils.constants import (CORE_ADDR, NEURONS1_ADDR, PCACHE_ADDR,
                                     PRIMARY_CACHE_STREAM_GROUP_ID)


class AllegoClient:
    """
    A python API for allego

    Attributes:
        samp_freq (float): current sampling rate
    """

    def __init__(self):
        """
        """

        self.samp_freq = None

    def restart(self, mode: str):
        """
        Restarts Allego in the requested mode.  Valid modes: 'sbpro', '='sbpro-sinaps-256', 'sbpro-sinaps-1024'
        'sbclassic', 'sim-sine', 'sim-spikes', 'open-ephys_usb2','open-ephys_usb3', 'intan1024', 'intan512',
        'xdaq-one-rec', 'xdaq-one-stim', 'xdaq-core-rec', 'xdaq-core-stim'

        Returns:
            None

        Side Effect:
            Resets all settings to their initialization values in the requested mode.

        Example:
            >>> client.restart('sim-spikes')
            None
        """
        return api.restart(CORE_ADDR, mode)

    def workspace_save(self, workspace_id=None, tags='', notes=''):
        """
        Saves current workspace

        Parameters:
            workspace_id (str): optional workspace ID
            tags (str): optional tags
            notes (str): optional notes

        Returns:
            None

        Example:
            >>> client.workspace_save(workspace_id='my_wspace', tags='my_tags', notes='my_notes)
            None
        """
        if workspace_id is None:
            return api.workspace_save(CORE_ADDR, True, tags, notes)
        else:
            return api.workspace_save_as(CORE_ADDR, workspace_id, True, tags, notes)

    def workspace_switch(self, workspace_id: str):
        """
        Switches to requested workspace

        Parameters:
            workspace_id (str): workspace ID

        Returns:
            None

        Example:
            >>> client.workspace_switch('my_wspace')
            None
        """
        return api.workspace_switch(CORE_ADDR, workspace_id)

    def workspace_delete(self, workspace_id: str):
        """
        Deletes requested workspace 

        Parameters:
            workspace_id (str): workspace ID

        Returns:
            None

        Example:
            >>> client.workspace_delete('my_wspace')
            None
        """
        return api.workspace_delete(CORE_ADDR, workspace_id)

    def workspace_current(self):
        """
        Returns current workspace ID

        Returns:
            workspaces (pandas.DataFrame)

        Example:
            >>> client.workspace_current()
            df
        """
        return api.workspace_current(CORE_ADDR)

    def workspace_list(self):
        """
        Returns table of all available workspaces

        Returns:
            workspaces (pandas.DataFrame)

        Example:
            >>> client.workspace_list()
            df
        """
        return api.workspace_list(CORE_ADDR)

    # ======= getters =======

    def get_connected_ports(self):
        """
        Gets dictionary of connected ports.

        Returns:
            ports (dict): dictionary with key=port label, value=number of channels.

        Example: 
            >>> client.get_connected_ports()
            {'A': 32, 'B': 256}

        See also:
            client.get_signal_group()

        """
        return api.get_ports(CORE_ADDR)

    def get_available_sensors(self):
        """
        Gets dictionary of dictionaries of available Radiens headstage and probe models, with each model of form key=component ID, value=number of channels. 

        Returns:
            sensors (dict): has keys 'headstages' and 'probes' that map to dictionaries of the available headstage models and probe models, respectively.  

        Example: 
            >>> client.get_available_sensors()
            {'sensors': {'headstages': {'hstg__chronic_smartlink_HC16': 16, 'hstg__acute_smartlink_A64': 64, ...}, 
                        'probes': {'a2x16_10mm50_500_177': 32, 'a1x16_5mm150_177': 16, ...}}
        """
        return api.get_sensors(CORE_ADDR)

    def get_status(self):
        """
        Gets the status of Allego.

        Returns:
            status (dict): has keys 'streaming' and 'recording' that map to dictionaries describing their respective status

        Example: 
            >>> client.get_status()
            {'recording': {'active_filename': 'allego_0__uid1020-00-53-02',
                        'duration': 0.0,
                        'error': '',
                        'mode': 'off'},
            'streaming': {'hardware_memory': 0.0, 'mode': 'off', 'time_range': [0.0, 0.0]}}
        """
        return api.get_status(CORE_ADDR)

    def get_config(self):
        """
        Gets the configuration of Allego.

        Returns:
            config (dict): has various keys relating to the system configuration

        Side Effect:
            sets ``samp_freq`` attribute to the current system sampling rate.

        Example:
            >>> client.get_config()
            {'backbone_mode': 'SMARTBOX_SIM_GEN_SPIKES',
            'base_samp_freq': 20000.0,
            'headstage_cable_lens': {'A': 3.0,
                                    'B': 3.0,
                                    'C': 3.0,
                                    'D': 3.0,
                                    'E': 3.0,
                                    'F': 3.0,
                                    'G': 3.0,
                                    'H': 3.0},
            'stream_loop_dur_ms': 200,
            'system_server_port': ':50051', 
            'base_filename': 'my_filename',
            'filepath': 'my_filepath' }

        """
        config = api.get_config(CORE_ADDR)
        self.samp_freq = config['base_samp_freq']

        return config

    def get_signal_group(self):
        """
        Gets the signal group (channel metadata) for all connected ports.

        Returns:
            signal_group (SignalGroup): object containing metadata information for all channels. See :ref:`SignalGroup <signal group>`

        Notes:
            All ports are scanned to detect the connected ports. 

        """
        return api.get_signal_group(CORE_ADDR, PRIMARY_CACHE_STREAM_GROUP_ID)

    def get_signals(self):
        """
        Gets the the most recent signals since last call of itself or :py:meth:`set_time_to_cache_head`.

        Returns:
            sigarray (ndarray[numpy.float32]): raw signal data; the mapping contained in signal_group (see :py:meth:`get_signal_group`)
            time_range (list[float]): time range (in seconds) of sigarray
        """
        if self.samp_freq is None:
            self.samp_freq = api.get_config(CORE_ADDR)['base_samp_freq']

        return api.get_signals(PCACHE_ADDR, PRIMARY_CACHE_STREAM_GROUP_ID, self.samp_freq)

    def get_digital_out_states(self):
        """
        Gets digital out state.

        Returns:
            dout_state_dict (dict): has keys 'digital_outs_mode' and 'states'.

        Example:
            >>> client.get_digital_out_states()
            {'digital_outs_mode': 'manual',
            'states': [{'chan_idx': 0, 'state': True}, {'chan_idx': 1, 'state': False}]}
        """
        return api.get_digital_out_states(CORE_ADDR)

    # ======= setters =======

    def set_samp_freq(self, fs: int):
        """
        Sets sampling frequency.  Valid sampling frequencies: 625, 1000, 1250, 1500, 2000, 2500, 3000, 3333,
        4000, 5000, 6250, 8000, 10000, 12500, 15000, 20000, 25000, 30000

        Parameters: 
            fs (int): requested sampling frequency (samples/sec)

        Returns:
            None

        Example:
            >>> client.set_samp_freq(12500)
            None

        """
        return api.set_fs(CORE_ADDR, fs)

    def set_sensor(self, port: str, headstage_id: str, probe_id: str):
        """
        Sets a sensor (headstage paired with probe) to the requested port.

        Parameters: 
            port (str): requested port (must be a currently connected port)
            headstage_id (str): requested headstage alias or headstage ID
            probe_id (str): requested probe alias or probe ID

        Returns:
            None

        Example:
            >>> client.set_sensor('A', 'smart-32', '4x8-32')
            None

        Notes:
            use radiens.utils.HEADSTAGE_ALIAS and radiens.utils.PROBE_ALIAS to get headstage and probe aliases, respectively. 

        See also:
            client.get_available_sensors()
            client.get_connected_ports()

        """
        return api.set_sensor(CORE_ADDR, port, headstage_id, probe_id)

    def set_time_to_cache_head(self):
        """
        Sets current time to most recent time point in signal cache.

        This should be called before the first call of :py:meth:`get_signals`. The first call of :py:meth:`get_signals` 
        will return the new signals since the last call to :py:meth:`set_time_to_cache_head` 
        """
        return api.set_time_to_cache_head(PCACHE_ADDR, PRIMARY_CACHE_STREAM_GROUP_ID)

    def set_streaming(self, mode: str):
        """
        Sets signal streaming to 'on' or 'off' 

        Parameters: 
            mode (str): requested mode, ``on`` or ``off``

        Returns:
            None

        Example:
            >>> client.set_streaming('on')
            None

            >>> client.set_streaming('off')
            None

        """
        if mode in ['on']:
            api.set_stream_state(CORE_ADDR, 1)
        elif mode in ['off']:
            api.set_stream_state(CORE_ADDR, 0)
        else:
            raise ValueError("mode must be in ['on', 'off', 1, 0, True, False]")

    def set_recording(self, mode: str):
        """
        Sets signal recording to 'on' or 'off' 

        Parameters: 
            mode (str): requested mode, ``on`` or ``off``

        Returns:
            None

        Side Effect:
            sets streaming on if `mode=``on`` and streaming is off

        Example:
            >>> client.set_recording('on')
            None

            >>> client.set_recording('off')
            None

        """
        if mode in ['on']:
            api.set_record_state(CORE_ADDR, 1)
        elif mode in ['off']:
            api.set_record_state(CORE_ADDR, 0)
        else:
            raise ValueError("mode must be in ['on', 'off', 1, 0, True, False]")

    def set_recording_config(self, datasource_path, datasource_name, datasource_idx, is_time_stamp=True):
        """
        Sets recording configuration

        Parameters: 
            datasource_path (str): path to output XDAT file set
            datasource_name (str): base name of XDAT file set
            datasource_idx (int): index number of output XDAT file set
            is_time_stamp (bool): `True` if file name includes time stamp

        Returns:
            None

        Side Effect:
            sets streaming on if `mode=``on`` and streaming is off

        Example:
            >>> client.set_recording('on')
            None

            >>> client.set_recording('off')
            None
        """
        if is_time_stamp in [None, True]:
            is_time_stamp = True
        elif is_time_stamp in [False]:
            is_time_stamp = False
        api.set_recording_config(CORE_ADDR, datasource_name, datasource_path, datasource_idx, is_time_stamp)

    def set_digital_out_manual(self, dout1_state: bool, dout2_state: bool):
        """
        Sets digital out state.

        Parameters:
            dout1_state (bool): new dout 1 state
            dout2_state (bool): new dout 2 state

        Returns:
            None

        Example:
            >>> client.set_digital_out_manual(true, false)
            None

        """
        return api.set_digital_out_manual(CORE_ADDR, dout1_state, dout2_state)

# ======= spike sorter =======

    def sorter_cmd(self, cmd: str):
        """
        Sends the requested command to the spike sorter.

        Available spike sorter commands: 'on', 'off', 'clear', 'rebase', 'init', 'localize', 'sort'

        Parameters:
            cmd (str): spike sorter command

        Returns:
            None

        Example:
            >>> client.sorter_cmd('on')
            None

        """
        return api.sorter_cmd(NEURONS1_ADDR, cmd)

    def sorter_get_params(self):
        """
        Returns the spike sorter parameters

        Returns:
            pandas.DataFrame

        Example:
            >>> client.sorter_get_params()
            None

        """
        return api.sorter_get_params(NEURONS1_ADDR)

    def sorter_get_state(self):
        """
        Returns the spike sorter state

        Returns:
            pandas.DataFrame

        Example:
            >>> client.sorter_get_state()
            None

        """
        return api.sorter_get_state(NEURONS1_ADDR)

    def sorter_get_dashboard(self):
        """
        Returns the spike sorter dashboard

        Returns:
            pandas.DataFrame

        Example:
            >>> client.sorter_get_dashboard()
            None

        """
        return api.sorter_get_dashboard(NEURONS1_ADDR)
